"""job_condition

Revision ID: 5a5c49313e52
Revises: 4b3c9d63e89c
Create Date: 2024-02-02 11:17:07.246330

"""
from alembic import op
import sqlalchemy as sa
import mindsdb.interfaces.storage.db  # noqa

# revision identifiers, used by Alembic.
revision = '5a5c49313e52'
down_revision = '4b3c9d63e89c'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('jobs', schema=None) as batch_op:
        batch_op.add_column(sa.Column('if_query_str', sa.String(), nullable=True))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('jobs', schema=None) as batch_op:
        batch_op.drop_column('if_query_str')

    # ### end Alembic commands ###
